# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Command line interface to interact with a manager for archives on the local
file system.
"""

import csv
import click
import pandas as pd
import sys

from histore.cli.archive import get_manager
from histore.document.base import InputDescriptor

import histore.util as util


# -- Create a new archive -----------------------------------------------------

@click.command(name='commit')
@click.option(
    '-b', '--basedir',
    required=False,
    type=click.Path(file_okay=False, dir_okay=True),
    help='Base directory for archive files'
)
@click.option(
    '-c', '--comment',
    required=False,
    type=str,
    help='Optional snapshot description'
)
@click.option(
    '-d', '--delimiter',
    required=False,
    type=str,
    help='One-character string used to separate fields'
)
@click.option(
    '-q', '--quotechar',
    required=False,
    type=str,
    help='One-character string used to quote fields with special characters'
)
@click.option(
    '-z', '--gzip',
    is_flag=True,
    default=False,
    help='Gzip compressed'
)
@click.argument(
    'archive',
    type=str
)
@click.argument(
    'filename',
    type=click.Path(file_okay=True, dir_okay=False, exists=True)
)
def commit_snapshot(
    basedir, comment, delimiter, quotechar, gzip, archive, filename
):
    """Commit file to archive."""
    store = get_archive(basedir, archive)
    if store is None:
        click.echo("Unknown archive '{}'".format(archive))
        sys.exit(-1)
    # Read the data frame.
    df = pd.read_csv(
        filename,
        delimiter=util.get_delimiter(delimiter),
        quotechar=quotechar if quotechar is not None else '"',
        quoting=csv.QUOTE_MINIMAL,
        compression='gzip' if gzip else None
    )
    s = store.commit(df, descriptor=InputDescriptor(description=comment))
    click.echo('Snapshot {} created.'.format(s.version))


# -- Checkout snapshot --------------------------------------------------------

@click.command(name='checkout')
@click.option(
    '-b', '--basedir',
    required=False,
    type=click.Path(file_okay=False, dir_okay=True),
    help='Base directory for archive files'
)
@click.option(
    '-d', '--delimiter',
    required=False,
    type=str,
    help='One-character string used to separate fields'
)
@click.option(
    '-q', '--quotechar',
    required=False,
    type=str,
    help='One-character string used to quote fields with special characters'
)
@click.option(
    '-z', '--gzip',
    is_flag=True,
    default=False,
    help='Gzip compressed'
)
@click.option(
    '-v', '--version',
    required=False,
    type=int,
    help='Snapshot version'
)
@click.argument(
    'archive',
    type=str
)
@click.argument(
    'filename',
    type=click.Path(file_okay=True, dir_okay=False)
)
def checkout_snapshot(
    basedir, delimiter, quotechar, gzip, version, archive, filename
):
    """Write snapshot to file."""
    store = get_archive(basedir, archive)
    if store is None:
        click.echo("Unknown archive '{}'".format(archive))
        sys.exit(-1)
    # Read the snapshot data frame.
    df = store.checkout(version=version)
    df.to_csv(
        filename,
        sep=util.get_delimiter(delimiter),
        quotechar=quotechar if quotechar is not None else '"',
        quoting=csv.QUOTE_MINIMAL,
        compression='gzip' if gzip else None,
        index=False
    )


# -- List snapshots -----------------------------------------------------------

@click.command(name='snapshots')
@click.option(
    '-b', '--basedir',
    required=False,
    type=click.Path(file_okay=False, dir_okay=True),
    help='Base directory for archive files'
)
@click.argument(
    'archive',
    type=str
)
def list_snapshots(basedir, archive):
    """List snapshots."""
    store = get_archive(basedir, archive)
    if store is None:
        click.echo("Unknown archive '{}'".format(archive))
        sys.exit(-1)
    for s in store.snapshots():
        click.echo('{}\t{}\t{}'.format(
            s.version,
            util.to_localtime(s.created_at).strftime('%Y-%m-%d %H:%M:%S'),
            s.description
        ))


# -- Helper Function ----------------------------------------------------------

def get_archive(basedir, name):
    """Get handle for archive with given name. Returns None if no archive with
    the given name exists.

    Parameters
    ----------
    basedir: string
        Base directory for the archive manager.
    name: string
        Unique archive name.

    Returns
    -------
    histore.archive.base.Archive
    """
    manager = get_manager(basedir)
    descriptor = manager.get_by_name(name)
    if descriptor is None:
        return None
    return manager.get(descriptor.identifier())



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Writer for archives that are materialized as Json files on the file
system.
"""

from typing import Optional

import json


from histore.archive.serialize.base import ArchiveSerializer
from histore.archive.serialize.default import DefaultSerializer
from histore.archive.row import ArchiveRow
from histore.archive.writer import ArchiveWriter
from histore.document.json.writer import JsonWriter


class ArchiveFileWriter(ArchiveWriter):
    """Archive writer that outputs rows in an archive as Json serialized rows
    in a text file. Each row is stored in a separate line in the text file. The
    output file is a Json array. The first and the last row of the file open
    and close the array.
    """
    def __init__(
        self, filename: str, row_counter: Optional[int] = 0,
        serializer: Optional[ArchiveSerializer] = None,
        compression: Optional[str] = None,
        encoder: Optional[json.JSONEncoder] = None
    ):
        """Initialize the output file, row counter, and the serializer that is
        being used.

        Parameters
        ----------
        filename: string
            Path to the output file.
        row_counter: int, default=0
            Counter that is used to generate unique internal row identifier.
            The current value of the counter is the value for the next unique
            identifier.
        serializer: histore.archive.serialize.base.ArchiveSerializer, default=None
            Implementation of the archive serializer interface that is used to
            serialize rows that are written to file.
        compression: string, default=None
            String representing the compression mode for the output file.
        encoder: json.JSONEncoder, default=None
            Encoder used when writing archive rows as JSON objects to file.
        """
        super(ArchiveFileWriter, self).__init__(row_counter)
        # Use the default serializer if no serializer was given
        self.serializer = serializer if serializer else DefaultSerializer()
        # Create Json writer for the archive rows.
        self.writer = JsonWriter(
            filename=filename,
            compression=compression,
            encoder=encoder
        )

    def close(self):
        """Write the last row to the output file and close the output array and
        the output file.
        """
        self.writer.close()

    def write_archive_row(self, row: ArchiveRow):
        """Add the given row to the output file.

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow
            Row in a new version of a dataset archive.
        """
        self.write_buffer(row)

    def write_buffer(self, row: Optional[ArchiveRow] = None):
        """Write the archive row in the internal buffer to the output file.
        Replace the buffer with the given (next output row).

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow, default=None
            Next row in the output stream. This row will be kept in the
            internal buffer and the previous row is being written to the
            output file.
        """
        self.writer.write(self.serializer.serialize_row(row))



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Reader for CSV files."""

from typing import Optional

import csv
import gzip

from histore.document.base import DataRow


class CSVReader(object):
    """Iterable reader for rows in a CSV file. Instances of this class act as
    an iterator for an open CSV file reader.
    """
    def __init__(
        self, filename: str, delim: Optional[str] = None, compressed: Optional[bool] = None,
        quotechar: Optional[str] = '"', encoding: Optional[str] = None,
        none_is: Optional[str] = None
    ):
        """Initialize the reader and the file handle.

        Parameters
        ----------
        filename: string
            Path to the input file.
        delim: string, default=None
            The column delimiter used in the CSV file.
        compressed: bool, default=None
            Flag indicating if the file contents are compressed using gzip.
        quotechar: string, default='"'
            CSV quote char.
        encoding: string, default=None
            The csv file encoding e.g. utf-8, utf-16 etc.
        none_is: string, default=None
            String that was used to encode None values in the input file. If
            given, all cell values that match the given string are substituted
            by None.
        """
        # Open file depending on whether it is gzip compressed or not.
        if compressed:
            self.file = gzip.open(filename, 'rt')
        else:
            self.file = open(filename, 'rt', encoding=encoding)
        # Open CSV reader.
        delim = delim if delim is not None else ','
        self.reader = csv.reader(self.file, delimiter=delim, quotechar=quotechar)
        self.none_is = none_is

    def __iter__(self):
        """Return object for row iteration."""
        return self

    def close(self):
        """Close the associated file handle and set it to None (to avoid
        repeated attempts to close an already closed file).
        """
        if self.file is not None:
            self.file.close()
            self.file = None
            self.buffer = None

    def __next__(self) -> DataRow:
        """Read the next row in the document.

        Returns
        -------
        histore.document.base.DataRow
        """
        # Read the next line into the buffer before returning the current
        # buffer result. Make sure to catch StopIteration errors if the end
        # of the file is reached.
        try:
            row = next(self.reader)
            if self.none_is is not None:
                row = [v if v != self.none_is else None for v in row]
            return row
        except StopIteration:
            self.close()
            raise StopIteration()



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Base classes for maintaining metadata about the archives that are controlled
by an archive manager.
"""

import json
import jsonschema

from datetime import datetime
from typing import Callable, Dict, List, Optional, Union

from histore.archive.serialize.base import ArchiveSerializer

import histore.util as util


"""Json schema for archive descriptors."""

DESCRIPTOR_SCHEMA = {
    'type': 'object',
    'properties': {
        'id': {'type': 'string'},
        'createdAt': {'type': 'string'},
        'name': {'type': 'string'},
        'description': {'type': 'string'},
        'primaryKey': {
            'type': 'array',
            'items': {'type': 'integer'}
        },
        'encoder': {'type': 'string'},
        'decoder': {'type': 'string'},
        'serializer': {
            'type': 'object',
            'properties': {
                'clspath': {'type': 'string'},
                'kwargs': {'type': 'object'}
            },
            'required': ['clspath']
        }
    },
    'required': ['id', 'createdAt']
}


class ArchiveDescriptor(object):
    """Wrapper around an archive descriptor dictionary object. This class
    provides access to descriptor property values and their defaults.
    """
    def __init__(self, doc: Dict, validate: Optional[bool] = True):
        """Initialize the dictionary containing the archive descriptor.
        Validates the document against the descriptor schema if the validate
        flag is True. Raises a ValidationError if validation fails.

        Parameters
        ----------
        doc: dict
            Dictionary containing an archive descriptor. The dictionary should
            adhere to the defined descriptor schema.
        validate: bool default=True
            Validate the given dictionary against the descriptor schema if this
            flag is True. Raises a ValidationError if the validation fails.

        Raises
        ------
        jsonschema.ValidationError
        """
        self.doc = doc
        if validate:
            jsonschema.validate(instance=doc, schema=DESCRIPTOR_SCHEMA)

    @staticmethod
    def create(
        identifier: Optional[str] = None, name: Optional[str] = None,
        primary_key: Optional[List[int]] = None, description: Optional[str] = None,
        encoder: Optional[str] = None, decoder: Optional[str] = None,
        serializer: Union[Dict, Callable] = None
    ):
        """Create a new archive descriptor object.

        Parameters
        ----------
        identifier: string, default=None
            Unique archive identifier.
        name: string, default=None
            Descriptive name that is associated with the archive.
        primary_key: list of int, default=None
            Identifier(s) for column(s) that are used to generate keys for rows
            in the archive.
        description: string, default=None
            Optional long description that is associated with the archive.
        encoder: string, default=None
            Full package path for the Json encoder class that is used by the
            persistent archive.
        decoder: string, default=None
            Full package path for the Json decoder function that is used by the
        serializer: dict or callable, default=None
            Dictionary or callable that returns a dictionary that contains the
            specification for the serializer. The serializer specification is
            a dictionary with the following elements:
            - ``clspath``: Full package target path for the serializer class
            that is instantiated.
            - ``kwargs`` : Additional arguments that are passed to the
            constructor of the created serializer instance.
            Only ``clspath`` is required.

        Returns
        -------
        histore.archive.manager.base.ArchiveDescriptor
        """
        # Create a unique identifier for the new archive.
        if identifier is None:
            identifier = util.get_unique_identifier()
        # Create the archive descriptor.
        doc = {'id': identifier, 'createdAt': util.current_time()}
        if name is not None:
            doc['name'] = name
        if description is not None:
            doc['description'] = description
        if primary_key is not None:
            doc['primaryKey'] = primary_key
        if encoder is not None:
            doc['encoder'] = encoder
        if decoder is not None:
            doc['decoder'] = decoder
        if serializer is not None:
            doc['serializer'] = serializer if isinstance(serializer, dict) else serializer()
        return ArchiveDescriptor(doc)

    def created_at(self) -> datetime:
        """Get creating timestamp for the archive.

        Returns
        -------
        datetime.datetime
        """
        return util.to_datetime(self.doc.get('createdAt'))

    def decoder(self) -> Callable:
        """Get Json decoder function used by persistent archives.

        Returns
        -------
        callable
        """
        return decoder_from_string(self.doc.get('decoder'))

    def description(self) -> str:
        """Get archive description. If the value is not set in the descriptor
        an empty string is returned as default.

        Returns
        -------
        string
        """
        return self.doc.get('description', '')

    def encoder(self) -> json.JSONEncoder:
        """Get Json encoder used by persistent archives.

        Returns
        -------
        json.JSONEncoder
        """
        return encoder_from_string(self.doc.get('encoder'))

    def identifier(self) -> str:
        """Get the unique archive identifier value.

        Returns
        -------
        string
        """
        return self.doc['id']

    def name(self) -> str:
        """Get the archive name. If the value is not set in the descriptor the
        identifier is returned as default.

        Returns
        -------
        string
        """
        return self.doc.get('name', self.identifier())

    def primary_key(self) -> List[int]:
        """Get list of primary key attribute identifier(s).

        Returns
        -------
        list of int
        """
        return self.doc.get('primaryKey')

    def rename(self, name: str):
        """Update the name of the archive.

        Parameters
        ----------
        name: string
            New archive name.
        """
        self.doc['name'] = name

    def serializer(self) -> ArchiveSerializer:
        """Get an instance of the serializer that is used for the archive.

        Returns
        -------
        histore.archive.serialize.base.ArchiveSerializer
        """
        return serializer_from_dict(self.doc.get('serializer'))


# -- Helper Functions ---------------------------------------------------------

def decoder_from_string(decoder: str) -> Callable:
    """Get Json decoder function from package path string.

    Parameters
    ----------
    decoder: string, default=None
        Full package path for the Json decoder function that is used by the
        persistent archive.

    Returns
    -------
    callable
    """
    return util.import_obj(decoder) if decoder else None


def encoder_from_string(encoder: str) -> json.JSONEncoder:
    """Get Json encoder from package path string.

    Parameters
    ----------
    encoder: string
        Full package path for the Json encoder class that is used by the
        persistent archive. May be None.

    Returns
    -------
    json.JSONEncoder
    """
    return util.import_obj(encoder) if encoder else None


def serializer_from_dict(doc: Dict) -> ArchiveSerializer:
    """Get an instance of an archive serializer from a dictionary serialization.

    Parameters
    ----------
    doc: dict
        Dictionary serialization for archive serializer. May be None.

    Returns
    -------
    histore.archive.serialize.base.ArchiveSerializer
    """
    return util.import_obj(doc['clspath'])(**doc.get('kwargs', {})) if doc is not None else None



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""The archive store intruduces a layer of abstraction between the archive
object and the way in which archives are managed and maintained.
"""

from abc import ABCMeta, abstractmethod
from typing import List

from histore.archive.reader import ArchiveReader
from histore.archive.schema import ArchiveSchema
from histore.archive.snapshot import SnapshotListing
from histore.archive.writer import ArchiveWriter


class ArchiveStore(metaclass=ABCMeta):
    """Abstract class that defines the interface to read and write dataset
    archive information.
    """
    @abstractmethod
    def commit(
        self, schema: ArchiveSchema, writer: ArchiveWriter,
        snapshots: SnapshotListing
    ):
        """Commit an updated dataset archive.

        The modified components of the archive are given as the three
        arguments of this method.

        Parameters
        ----------
        schema: histore.archive.schema.ArchiveSchema
            Schema history for the new archive version.
        writer: histore.archive.writer.ArchiveWriter
            Instance of the archive writer class returned by this store that
            was used to output the rows of the new archive version.
        snapshots: histore.archive.snapshot.SnapshotListing
            Snapshot listing for the modified archive.
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def is_empty(self) -> bool:
        """True if the archive does not contain any snapshots yet.

        Returns
        -------
        bool
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_reader(self) -> ArchiveReader:
        """Get the row reader for this archive.

        Returns
        -------
        histore.archive.reader.ArchiveReader
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_schema(self) -> ArchiveSchema:
        """Get the schema history for the archived dataset.

        Returns
        -------
        histore.archive.schema.ArchiveSchema
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_snapshots(self) -> SnapshotListing:
        """Get listing of all snapshots in the archive.

        Returns
        -------
        histore.archive.snapshot.SnapshotListing
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_writer(self) -> ArchiveWriter:
        """Get a writer for a new version of the archive.

        Returns
        -------
        histore.archive.writer.ArchiveWriter
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def primary_key(self) -> List[int]:
        """Get the list of identifier for the primary key column(s).

        Returns None if the archive is not keyed by a primary key.

        Returns
        -------
        list of int
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def rollback(self, schema: ArchiveSchema, writer: ArchiveWriter, version: int):
        """Store the archive after it has been rolled back to a previous
        version.

        The archive schema and archive writer contain the modified schema and
        rows after the rollback.

        Parameters
        ----------
        schema: histore.archive.schema.ArchiveSchema
            Schema history for the previous archive version.
        writer: histore.archive.writer.ArchiveWriter
            Instance of the archive writer class returned by this store that
            was used to output the rows of the previous archive version.
        version: int
            Unique version identifier for the rollback snapshot.
        """
        raise NotImplementedError()  # pragma: no cover



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Manager for archives that maintains all archive metadata in a relational
database.
"""

from sqlalchemy.exc import SQLAlchemyError
from typing import Callable, Dict, Optional, Union

import os
import shutil

from histore.archive.base import InputDocument, PrimaryKey, get_key_columns, to_document
from histore.archive.base import Archive as PersistentArchive
from histore.archive.manager.base import ArchiveManager
from histore.archive.manager.db.database import DB
from histore.archive.manager.db.model import Archive, ArchiveKey
from histore.archive.manager.descriptor import ArchiveDescriptor
from histore.archive.manager.descriptor import decoder_from_string, encoder_from_string, serializer_from_dict
from histore.archive.store.fs.base import ArchiveFileStore
from histore.document.base import InputDescriptor

import histore.config as config
import histore.util as util


class DBArchiveManager(ArchiveManager):
    """The database archive manager maintains a set of archives on disk. The
    list of archive descriptors is maintained in a relational database.
    """
    def __init__(
        self, basedir: Optional[str] = None, db: Optional[DB] = None,
        create: Optional[bool] = False
    ):
        """Initialize the base directory under which all archives are stored in
        individual sub-folders and the database connection object for archive
        descriptors.

        If the base directory is not given the value will be read from the
        environment variable HISTORE_BASEDIR. If the database is not given
        the connection string will be read from the environment variable
        HISTORE_DBCONNECT.

        Parameters
        ----------
        basedir: string, default=None
            Path to dorectory on disk where archives are maintained.
        db: histore.archive.manager.db.database.DB, default=None
            Database connection object.
        create: bool, default=False
            Create a fresh database and delete all files in the base directory
            if True.
        """
        # Initialize the base directory.
        if basedir is None:
            basedir = config.BASEDIR()
        self.basedir = basedir
        if create:
            # Clear all files in the base directory if it exists.
            if os.path.isdir(self.basedir):
                util.cleardir(self.basedir)
        # Initialize the database connector.
        if db is None:
            db = DB(connect_url=config.DBCONNECT())
        self.db = db
        if create:
            # Create a fresh database instance.
            db.init()
        # Create the base directory (if it does not exist).
        util.createdir(basedir)

    def _archive_dir(self, identifier: str) -> str:
        """Get directory for archive with the given identifier.

        Parameters
        ----------
        identifier: string
            Unique archive identifier.

        Returns
        -------
        string
        """
        return os.path.join(self.basedir, identifier)

    def archives(self) -> Dict[str, ArchiveDescriptor]:
        """Get dictionary of archive descriptors. The returned dictionary maps
        archive identifier to their descriptor.

        Returns
        -------
        dict(string: histore.archive.manager.descriptor.ArchiveDescriptor)
        """
        archives = dict()
        with self.db.session() as session:
            for arch in session.query(Archive).all():
                archives[arch.archive_id] = arch.descriptor()
        return archives

    def create(
        self, name: Optional[str] = None, description: Optional[str] = None,
        encoder: Optional[str] = None, decoder: Optional[str] = None,
        serializer: Union[Dict, Callable] = None, doc: Optional[InputDocument] = None,
        primary_key: Optional[PrimaryKey] = None, snapshot: Optional[InputDescriptor] = None,
        sorted: Optional[bool] = False, buffersize: Optional[float] = None,
        validate: Optional[bool] = False
    ) -> ArchiveDescriptor:
        """Create a new archive object under a given unique name.

        For archives that are keyed by a primary key, the input document for
        the first dataset snapshot  has to be provided. This snapshot will be
        loaded into the archive.

        Raises a ValueError if an archive with the given name exists.

        Parameters
        ----------
        name: string, default=None
            Descriptive name that is associated with the archive.
        description: string, default=None
            Optional long description that is associated with the archive.
        encoder: string, default=None
            Full package path for the Json encoder class that is used by the
            persistent archive.
        decoder: string, default=None
            Full package path for the Json decoder function that is used by the
            persistent archive.
        serializer: dict or callable, default=None
            Dictionary or callable that returns a dictionary that contains the
            specification for the serializer. The serializer specification is
            a dictionary with the following elements:
            - ``clspath``: Full package target path for the serializer class
            that is instantiated.
            - ``kwargs`` : Additional arguments that are passed to the
            constructor of the created serializer instance.
            Only ``clspath`` is required.
        doc: histore.archive.base.InputDocument, default=None
            Input document representing the initial dataset snapshot that is
            being loaded into the archive.
        primary_key: string or list, default=None
            Column(s) that are used to generate identifier for snapshot rows.
        snapshot: histore.document.base.InputDescriptor, default=None
            Optional metadata for the created snapshot.
        sorted: bool, default=False
            Flag indicating if the document is sorted by the optional primary
            key attributes. Ignored if the archive is not keyed.
        buffersize: float, default=None
            Maximum size (in bytes) for the memory buffer when sorting the
            input document.
        validate: bool, default=False
            Validate that the resulting archive is in proper order before
            committing the action.

        Returns
        -------
        histore.archive.manager.descriptor.ArchiveDescriptor

        Raises
        ------
        ValueError
        """
        # Get serializer dictionary if a function was given.
        serializer = serializer() if serializer is not None and callable(serializer) else serializer
        # Create an unique identifier for the new archive.
        identifier = util.get_unique_identifier()
        try:
            with self.db.session() as session:
                archive = Archive(
                    archive_id=identifier,
                    name=name,
                    description=description,
                    encoder=encoder,
                    decoder=decoder,
                    serializer=serializer
                )
                # Load initial snapshot if given.
                if doc is not None:
                    doc = to_document(doc)
                    # Get the expected identifier for the primary key columns.
                    primary_key = get_key_columns(columns=doc.columns, primary_key=primary_key)
                    store = ArchiveFileStore(
                        basedir=self._archive_dir(identifier),
                        replace=True,
                        serializer=serializer_from_dict(serializer),
                        encoder=encoder_from_string(encoder),
                        decoder=decoder_from_string(decoder),
                        primary_key=primary_key
                    )
                    arch = PersistentArchive(store=store)
                    arch.commit(
                        doc=doc,
                        descriptor=snapshot,
                        sorted=sorted,
                        buffersize=buffersize,
                        validate=validate
                    )
                elif primary_key is not None:
                    raise ValueError('missing snapshot document')
                if primary_key is not None:
                    for pos, colid in enumerate(primary_key):
                        key = ArchiveKey(colid=colid, pos=pos)
                        archive.keyspec.append(key)
                session.add(archive)
                session.commit()
                return archive.descriptor()
        except SQLAlchemyError as ex:
            raise ValueError(ex)

    def delete(self, identifier: str):
        """Delete the archive with the given identifier.

        Parameters
        ----------
        identifier: string
            Unique archive identifier
        """
        with self.db.session() as session:
            # Query database to ensure that the archive exists.
            archive = session\
                .query(Archive)\
                .filter(Archive.archive_id == identifier)\
                .one_or_none()
            if archive is None:
                return
            # Remove the archive base directory and the entry in the database.
            archdir = self._archive_dir(identifier)
            shutil.rmtree(archdir)
            session.delete(archive)

    def get(self, identifier: str) -> PersistentArchive:
        """Get the archive that is associated with the given identifier. Raises
        a ValueError if the identifier is unknown.

        Parameters
        ----------
        identifier: string
            Unique archive identifier

        Returns
        -------
        histore.archive.base.Archive

        Raises
        ------
        ValueError
        """
        with self.db.session() as session:
            # Query database to ensure that the archive exists.
            archive = session\
                .query(Archive)\
                .filter(Archive.archive_id == identifier)\
                .one_or_none()
            if archive is None:
                raise ValueError('unknown archive {}'.format(identifier))
            # Get the archive descriptor and close the database connection.
            desc = archive.descriptor()
        store = ArchiveFileStore(
            basedir=self._archive_dir(identifier),
            replace=False,
            serializer=desc.serializer(),
            encoder=desc.encoder(),
            decoder=desc.decoder(),
            primary_key=desc.primary_key()
        )
        return PersistentArchive(store=store)

    def rename(self, identifier: str, name: str):
        """Rename the specified archive. Raises a ValueError if the identifier
        is unknown or if an archive with the given name exist.

        Parameters
        ----------
        identifier: string
            Unique archive identifier
        name: string
            New archive name.

        Raises
        ------
        ValueError
        """
        try:
            with self.db.session() as session:
                # Query database to ensure that the archive exists.
                archive = session\
                    .query(Archive)\
                    .filter(Archive.archive_id == identifier)\
                    .one_or_none()
                if archive is None:
                    raise ValueError('unknown archive {}'.format(identifier))
                # Update the archive name in the database.
                archive.name = name
        except SQLAlchemyError as ex:
            raise ValueError(ex)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Archives are collections of snapshots of an evolving dataset."""

from collections import defaultdict
from typing import Callable, Dict, List, Optional, Tuple, Union

import pandas as pd

from histore.archive.manager.descriptor import decoder_from_string, encoder_from_string, serializer_from_dict
from histore.archive.provenance.archive import SnapshotDiff
from histore.archive.reader import ArchiveReader, SnapshotReader
from histore.archive.schema import ArchiveSchema
from histore.archive.snapshot import Snapshot, SnapshotListing
from histore.archive.store.base import ArchiveStore
from histore.archive.store.fs.base import ArchiveFileStore
from histore.archive.store.mem.base import VolatileArchiveStore
from histore.archive.writer import ValidatingArchiveWriter
from histore.document.base import Document, InputDescriptor
from histore.document.csv.base import CSVFile
from histore.document.df import DataFrameDocument
from histore.document.operator import DatasetOperator
from histore.document.reader import AnnotatedDocumentReader, DefaultDocumentReader
from histore.document.schema import DocumentSchema

import histore.archive.merge as nested_merge


"""Type aliases for archive API methods."""
InputDocument = Union[pd.DataFrame, str, Document]
# Primary key of a dataset.
PrimaryKey = Union[str, List[str]]


class Archive(object):
    """An archive maintains a list of dataset snapshots. Archive snapshots are
    maintained by an archive store. The store provides a layer of abstraction
    such that the archive object does not have to deal with the different ways
    in which archives are managed by different systems.
    """
    def __init__(
        self, store: Optional[ArchiveStore] = None, doc: Optional[InputDocument] = None,
        primary_key: Optional[PrimaryKey] = None, descriptor: Optional[InputDescriptor] = None,
        sorted: Optional[bool] = False, buffersize: Optional[float] = None,
        validate: Optional[bool] = False
    ):
        """Initialize the associated archive store.

        If the archive store is None a volatile store is used as the default. In
        this case the user can provide informaiton for a first document (and its
        primary key) that will be committed into the new archive.

        Parameters
        ----------
        store: histore.archive.store.base.ArchiveStore, default=None
            Associated archive store that is used to maintain all archive
            information. Uses the volatile in-memory store by default.
        doc: histore.archive.base.InputDocument, default=None
            Input document representing the initial dataset snapshot that is
            being loaded into the archive.
        primary_key: string or list, default=None
            Column(s) that are used to generate identifier for snapshot rows.
        descriptor: histore.document.base.InputDescriptor, default=None
            Optional metadata for the created snapshot.
        sorted: bool, default=False
            Flag indicating if the document is sorted by the optional primary
            key attributes. Ignored if the archive is not keyed.
        buffersize: float, default=None
            Maximum size (in bytes) for the memory buffer when sorting the
            input document.
        validate: bool, default=False
            Validate that the resulting archive is in proper order before
            committing the action.
        """
        if store is not None:
            # If a archive store is given the user cannot provide a document with
            # a primary key.
            self.store = store
            if doc is not None:
                if primary_key is not None:
                    raise ValueError('cannot load initial dataset with primary key')
                # Load the document.
                self.commit(
                    doc=doc,
                    descriptor=descriptor,
                    sorted=sorted,
                    buffersize=buffersize,
                    validate=validate
                )
        else:
            # Create a volatile archive store. If a document is provided this
            # will be committed as the first snapshot.
            if doc is not None:
                doc = to_document(doc)
                # Get the expected identifier for the primary key columns.
                primary_key = get_key_columns(columns=doc.columns, primary_key=primary_key)
                store = VolatileArchiveStore(primary_key=primary_key)
                archive = Archive(store=store)
                archive.commit(
                    doc=doc,
                    descriptor=descriptor,
                    sorted=sorted,
                    buffersize=buffersize,
                    validate=validate
                )
            elif primary_key is not None:
                raise ValueError('missing snapshot document')
            else:
                store = VolatileArchiveStore()
            self.store = store

    def apply(
        self, operators: Union[DatasetOperator, List[DatasetOperator]],
        origin: Optional[int] = None, validate: Optional[bool] = None
    ) -> List[Snapshot]:
        """Apply a given operator or a sequence of operators on a snapshot in
        the archive.

        The resulting snapshot(s) will directly be merged into the archive. This
        method allows to update data in an archive directly without the need
        to checkout the snapshot first and then commit the modified version(s).

        Returns list of handles for the created snapshots.

        Note that there are some limitations for this method. Most importantly,
        the order of rows cannot be modified and neither can it insert new rows
        at this point. Columns can be added, moved, renamed, and deleted.
        It is also important to note that if rows are deleted the position
        information in the created snapshot will not reflect this accurately.
        That is, the ordering of rows will still be correct, but the positions
        are no longer absolute positions since some positin values may be
        missing.

        Parameters
        ----------
        operators: histore.document.operator.DatasetOperator or
                list of histore.document.stream.DatasetOperator
            Operator(s) that is/are used to update the rows in a dataset
            snapshot to create new snapshot(s) in this archive.
        origin: int, default=None
            Unique version identifier for the original snapshot that is being
            updated. By default the last version is updated.
        validate: bool, default=False
            Validate that the resulting archive is in proper order before
            committing the action.

        Returns
        -------
        histore.archive.snapshot.Snapshot
        """
        # Ensure that the origin version is set.
        origin = origin if origin is not None else self.snapshots().last_snapshot().version
        # Create pipeline of operators. Make sure to update the archive schema
        # according to the operators. Each operator will create a new snapshot
        # version.
        result = list()
        pipeline = list()
        schema = self.schema()
        origin_colids = [c.colid for c in schema.at_version(version=origin)]
        # Get snapshot identifier and schema information for each created
        # snapshot.
        snapshots = self.snapshots()
        for op in operators if isinstance(operators, list) else [operators]:
            version = snapshots.next_version()
            snapshots = snapshots.append(version=version, descriptor=op)
            result.append(snapshots.last_snapshot())
            # Merge the new snapshot schema with the current archive schema. Then
            # get the snapshot schema from the merged archive schema to ensure we
            # have all the column identifier.
            schema, columns = schema.merge(
                columns=op.columns,
                version=version,
                origin=origin
            )
            pipeline.append((op, version, columns))
        # Iterate over rows and apply the operator to those that belong to the
        # modified snapshot.
        writer = self.store.get_writer()
        if validate:
            writer = ValidatingArchiveWriter(writer=writer)
        with self.reader() as reader:
            for row in reader:
                if row.timestamp.contains(origin):
                    pos, vals = row.at_version(version=origin, columns=origin_colids)
                    for op, version, snapshot_colids in pipeline:
                        vals = op.handle(rowid=row.rowid, row=vals)
                        if vals is not None:
                            # Merge the archive row and the modified document row.
                            row = row.merge(
                                values={colid: val for colid, val in zip(snapshot_colids, vals)},
                                pos=pos,
                                version=version
                            )
                        else:
                            break
                writer.write_archive_row(row)
        # Commit all changes to the associated archive store. Make sure to
        # unwrapt the writer if validating.
        if validate:
            writer = writer.writer
        self.store.commit(
            schema=schema,
            writer=writer,
            snapshots=snapshots
        )
        # Return descriptor for the created snapshot.
        return result

    def checkout(self, version: Optional[int] = None) -> pd.DataFrame:
        """Access a dataset snapshot in the archive.

        Retrieves the datset that was commited with the given version
        identifier. Raises an error if the version identifier is unknown. If no
        version identifier is given the last snapshot will be returned by
        default.

        Parameters
        ----------
        version: int, default=None
            Unique version identifier. By default the last version is returned.

        Returns
        -------
        pandas.DataFrame

        Raises
        ------
        ValueError
        """
        # Use the last snapshot as default if no version is specified. Ensure
        # a given version exists in the snapshot index.
        if version is None:
            version = self.snapshots().last_snapshot().version
        elif not self.snapshots().has_version(version):
            raise ValueError('unknown version {}'.format(version))
        # The index values for the returned data frame depend on the the type
        # of key that was used for the archive. If the archive does not have a
        # primary key, the row key represents the row index value. If a primary
        # key is defined, we use the row identifier as the row index value.
        is_keyed = self.store.primary_key() is not None
        # Get dataset schema at the given version.
        columns = self.schema().at_version(version)
        colids = [c.colid for c in columns]
        # Get the row values and their position.
        rows = list()
        with self.reader() as reader:
            for row in reader:
                if row.timestamp.contains(version):
                    pos, vals = row.at_version(version, colids, raise_error=False)
                    rowidx = row.rowid if is_keyed else row.key.value
                    rows.append((rowidx, pos, vals))
        # Sort rows in ascending order.
        rows.sort(key=lambda r: r[1])
        # Create document for the retrieved snapshot.
        data, rowindex = list(), list()
        for rowid, _, vals in rows:
            data.append(vals)
            rowindex.append(rowid)
        return pd.DataFrame(data=data, index=rowindex, columns=columns, dtype=object)

    def commit(
        self, doc: InputDocument, descriptor: Optional[InputDescriptor] = None,
        sorted: Optional[bool] = False, buffersize: Optional[float] = None,
        validate: Optional[bool] = None, renamed: Optional[List[Tuple[str, str]]] = None
    ) -> Snapshot:
        """Commit a new snapshot to the dataset archive.

        The given document represents the dataset snapshot that is being merged
        into the archive. Returns the descriptor of the merged snapshot in the
        new version of the archive.

        Parameters
        ----------
        doc: histore.archive.base.InputDocument, or string
            Input document representing the dataset snapshot that is being
            merged into the archive.
        descriptor: histore.document.base.InputDescriptor, default=None
            Optional metadata for the created snapshot.
        sorted: bool, default=False
            Flag indicating if the document is sorted by the optional primary
            key attributes. Ignored if the archive is not keyed.
        buffersize: float, default=None
            Maximum size (in bytes) for the memory buffer when sorting the
            input document.
        validate: bool, default=False
            Validate that the resulting archive is in proper order before
            committing the action.
        renamed: list of tuple of str, str, default=None
            Optional mapping of columns that have been renamed. Tuples contain
            the old and the new columns name.

        Returns
        -------
        histore.archive.snapshot.Snapshot
        """
        # Ensure that we have an instance of the Document class as input.
        doc = to_document(doc=doc)
        # Get the current snapshot version. If the archive is empty the origin
        # is None
        origin = self.snapshots().last_snapshot().version if not self.is_empty() else None
        try:
            # Get version identifier for the next snapshot.
            version = self.snapshots().next_version()
            # Merge the new snapshot schema with the current archive schema.
            schema, columns = self.schema().merge(
                columns=doc.columns,
                version=version,
                renamed=renamed,
                origin=origin
            )
            # Sort the document if the archive has a primary key and the
            # document is not sorted.
            if self.store.primary_key() and not sorted:
                mapping = {colid: colidx for colidx, colid in enumerate(columns)}
                doc = doc.sorted(
                    keys=[mapping[colid] for colid in self.store.primary_key()],
                    buffersize=buffersize
                )
            # Create the document reader. The type of document reader depends
            # on the type of key that is used for the document.
            with doc.open() as iterator:
                if self.store.primary_key():
                    reader = AnnotatedDocumentReader(
                        iterator=iterator,
                        columns=columns,
                        keys=self.store.primary_key()
                    )
                else:
                    reader = DefaultDocumentReader(iterator=iterator, columns=columns)
                # Merge document rows into the archive.
                validate = validate if validate is not None else sorted
                writer = self.store.get_writer()
                if validate:
                    writer = ValidatingArchiveWriter(writer=writer)
                nested_merge.merge_rows(
                    arch_reader=self.reader(),
                    doc_reader=reader,
                    version=version,
                    writer=writer
                )
        finally:
            # Ensure that the close method for the document is called to allow
            # for cleanup of temporary files.
            doc.close()
        # Commit all changes to the associated archive store. Make sure to
        # unwrapt the writer if validating.
        if validate:
            writer = writer.writer
        self.store.commit(
            schema=schema,
            writer=writer,
            snapshots=self.snapshots().append(version=version, descriptor=descriptor)
        )
        # Return descriptor for the created snapshot.
        return self.snapshots().last_snapshot()

    def diff(self, original_version: int, new_version: int) -> SnapshotDiff:
        """Get provenance information representing the difference between two
        dataset snapshots.

        Parameters
        ----------
        original_version: int
            Unique identifier for the original version.
        new_version: int
            Unique identifier for the version that the original version is
            compared against.

        Returns
        -------
        histore.archive.provenance.archive.SnapshotDiff
        """
        prov = SnapshotDiff()
        # Get changes in the dataset schema.
        schema_diff = prov.schema()
        for colid, column in self.schema().columns.items():
            col_prov = column.diff(original_version, new_version)
            if col_prov is not None:
                schema_diff.add(col_prov)
        # Get changes in dataset rows.
        rows_diff = prov.rows()
        row_count = 0
        with self.reader() as reader:
            for row in reader:
                row_prov = row.diff(original_version, new_version)
                if row_prov is not None:
                    rows_diff.add(row_prov)
                row_count += 1
        return prov

    def is_empty(self) -> bool:
        """True if the archive does not contain any snapshots yet.

        Returns
        -------
        bool
        """
        return self.store.is_empty()

    def open(self, version: Optional[int] = None) -> SnapshotReader:
        """Get a stream reader for a dataset snapshot.

        Parameters
        ----------
        version: int, default=None
            Unique version identifier. By default the last version is used.

        Returns
        -------
        histore.archive.reader.SnapshotReader
        """
        # Use the last snapshot as default if no version is specified.
        if version is None:
            version = self.snapshots().last_snapshot().version
        # Ensure that the version exists in the snapshot index.
        if not self.snapshots().has_version(version):
            raise ValueError('unknown version {}'.format(version))
        # Get dataset schema at the given version.
        columns = self.schema().at_version(version)
        return SnapshotReader(
            reader=self.reader,
            version=version,
            schema=columns,
            is_keyed=self.store.primary_key() is not None
        )

    def reader(self) -> ArchiveReader:
        """Get the row reader for this archive.

        Returns
        -------
        histore.archive.reader.ArchiveReader
        """
        return self.store.get_reader()

    def rollback(self, version: int):
        """Rollback the archive history to the snapshot with the given version
        identifier.

        Raises a ValueError if the given version identifier is invalid.

        Parameters
        ----------
        version: int
            Unique version identifier for the last version in the archive after
            rollback.

        Raises
        ------
        ValueError
        """
        # Ensure that the given version identifier is valid.
        if not self.snapshots().has_version(version=version):
            raise ValueError("unknown version '{}'".format(version))
        # Rollback archive rows.
        reader = self.reader()
        writer = self.store.get_writer()
        for row in reader:
            row = row.rollback(version)
            if row is not None:
                writer.write_archive_row(row)
        # Commit the rolledback archive to the associated archive store.
        self.store.rollback(
            schema=self.schema().rollback(version=version),
            writer=writer,
            version=version
        )

    def schema(self) -> ArchiveSchema:
        """Get the schema history for the archived dataset.

        Returns
        -------
        histore.archive.schema.ArchiveSchema
        """
        return self.store.get_schema()

    def snapshots(self) -> SnapshotListing:
        """Get listing of all snapshots in the archive.

        Returns
        -------
        histore.archive.snapshot.SnapshotListing
        """
        return self.store.get_snapshots()


class PersistentArchive(Archive):
    """Archive that persists all dataset snapshots in files on the file system.
    All the data is maintained within a given directory on the file system.
    This class is a wrapper for an archive with a persistent archive store. It
    uses the default file system store.
    """
    def __init__(
        self, store: Optional[ArchiveStore] = None, basedir: Optional[str] = None,
        create: Optional[bool] = False, encoder: Optional[str] = None,
        decoder: Optional[str] = None, serializer: Union[Dict, Callable] = None,
        doc: Optional[InputDocument] = None, primary_key: Optional[PrimaryKey] = None,
        descriptor: Optional[InputDescriptor] = None, sorted: Optional[bool] = False,
        buffersize: Optional[float] = None, validate: Optional[bool] = False
    ):
        """Initialize the associated archive store.

        If the archive store is None a volatile store is used as the default. In
        this case the user can provide informaiton for a first document (and its
        primary key) that will be committed into the new archive.

        Parameters
        ----------
        store: histore.archive.store.base.ArchiveStore, default=None
            Associated archive store that is used to maintain all archive
            information. Uses the volatile in-memory store by default.
        basedir: string
            Path to the base directory for archive files. If the directory does
            not exist it will be created.
        create: boolean, default=False
            Do not read existing files in the base directory to initialize the
            archive. Treat the archive as being empty instead if True.
        encoder: string, default=None
            Full package path for the Json encoder class that is used by the
            persistent archive.
        decoder: string, default=None
            Full package path for the Json decoder function that is used by the
            persistent archive.
        serializer: dict or callable, default=None
            Dictionary or callable that returns a dictionary that contains the
            specification for the serializer. The serializer specification is
            a dictionary with the following elements:
            - ``clspath``: Full package target path for the serializer class
                           that is instantiated.
            - ``kwargs`` : Additional arguments that are passed to the
                           constructor of the created serializer instance.
            Only ``clspath`` is required.
        doc: histore.archive.base.InputDocument, default=None
            Input document representing the initial dataset snapshot that is
            being loaded into the archive.
        primary_key: string or list, default=None
            Column(s) that are used to generate identifier for snapshot rows.
        descriptor: histore.document.base.InputDescriptor, default=None
            Optional metadata for the created snapshot.
        sorted: bool, default=False
            Flag indicating if the document is sorted by the optional primary
            key attributes. Ignored if the archive is not keyed.
        buffersize: float, default=None
            Maximum size (in bytes) for the memory buffer when sorting the
            input document.
        validate: bool, default=False
            Validate that the resulting archive is in proper order before
            committing the action.
        """
        if store is not None:
            # If a archive store is given the user cannot provide a document with
            # a primary key.
            self.store = store
            if doc is not None:
                if primary_key is not None:
                    raise ValueError('cannot load initial dataset with primary key')
                # Load the document.
                self.commit(
                    doc=doc,
                    descriptor=descriptor,
                    sorted=sorted,
                    buffersize=buffersize,
                    validate=validate
                )
        else:
            # Create a persistent archive store. If a document is provided this
            # will be committed as the first snapshot.
            if doc is not None:
                doc = to_document(doc)
                # Get the expected identifier for the primary key columns.
                primary_key = get_key_columns(columns=doc.columns, primary_key=primary_key)
                # Get serializer dictionary if a function was given.
                serializer = serializer() if serializer is not None and callable(serializer) else serializer
                store = ArchiveFileStore(
                    basedir=basedir,
                    replace=create,
                    serializer=serializer_from_dict(serializer),
                    encoder=encoder_from_string(encoder),
                    decoder=decoder_from_string(decoder),
                    primary_key=primary_key
                )
                archive = Archive(store=store)
                archive.commit(
                    doc=doc,
                    descriptor=descriptor,
                    sorted=sorted,
                    buffersize=buffersize,
                    validate=validate
                )
            elif primary_key is not None:
                raise ValueError('missing snapshot document')
            else:
                store = ArchiveFileStore(
                    basedir=basedir,
                    replace=create,
                    serializer=serializer_from_dict(serializer),
                    encoder=encoder_from_string(encoder),
                    decoder=decoder_from_string(decoder)
                )
            self.store = store


# -- Helper Functions ---------------------------------------------------------

def get_key_columns(columns: DocumentSchema, primary_key: PrimaryKey) -> List[int]:
    """Get identifier for primary key columns.

    Uses the archive schema merge method to get the identifiable columns for the
    given schema. Then returns the identifier for the columns that match the
    primary key columns.

    Note that the primary key specification cannot reference columns that have
    non-unique names. If a primary key column occurs more than once in the
    resulting schema an error is raised. An error is also raised if the primary
    key references a column that does not exist.

    If the list of primary key columns is None or empty the result is None.

    Parameters
    ----------
    columns: list of string
        Schema for the input document.
    primary_key: list of string
        Names of primary key columns. May be None.

    Returns
    -------
    list of int
    """
    # Return None if the list of primary key columns is None or empty.
    if not primary_key:
        return None
    # Ensure that the primary key is a list.
    primary_key = primary_key if isinstance(primary_key, list) else [primary_key]
    # Get the identifiable columns for the document schema. Create a mapping
    # from the column name to the identifier(s).
    schema, _ = ArchiveSchema().merge(columns=columns, version=0)
    columns = defaultdict(list)
    for col in schema.at_version(0):
        columns[col].append(col.colid)
    # Generate the list of primary key column identifier.
    pk = list()
    for name in primary_key:
        col = columns.get(name)
        if not col:
            raise ValueError("unknown column '{}'".format(col))
        elif len(col) > 1:
            raise ValueError("not a unique key column '{}'".format(col))
        pk.append(col[0])
    return pk


def to_document(doc: InputDocument) -> Document:
    """Ensure that a given input document is an instance of class Document.

    For file names and pandas data frame inputs the repsective document is
    returned.

    Parameters
    ----------
    doc: histore.document.base.Document, pd.DataFrame,
            histore.document.csv.base.CSVFile, or string
        Input document representing a dataset snapshot.

    Returns
    -------
    histore.document.base.Document
    """
    if isinstance(doc, pd.DataFrame):
        # If the given document is a pandas DataFrame we need to wrap it in
        # the appropriate document class.
        return DataFrameDocument(df=doc)
    elif isinstance(doc, str):
        return CSVFile(filename=doc)
    return doc



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

from abc import ABCMeta, abstractmethod
from typing import Callable, List, Tuple

from histore.document.base import DefaultDocument, DocumentIterator, RowIndex, DataRow
from histore.document.schema import Column


class ArchiveReader(metaclass=ABCMeta):
    """Reader for rows in a dataset archive. Reads rows in ascending order of
    their identifier.
    """
    def __enter__(self) -> DocumentIterator:
        """Enter method for the context manager."""
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        """Close the document iterator when the context manager exits."""
        self.close()
        return False

    def __iter__(self):
        """Make the reader instance iterable by returning a generator that
        yields all rows.

        Returns
        -------
        Generator
        """
        return row_stream(self)

    @abstractmethod
    def close(self):
        """Release all reseources that are associated with the reader."""
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def next(self):
        """Read the next row in the dataset archive.

        Returns None if the end of the archive rows has been reached.

        Returns
        -------
        histore.archive.row.ArchiveRow
        """
        raise NotImplementedError()  # pragma: no cover


class SnapshotIterator(DocumentIterator):
    """Reader for data streams. Provides the functionality to open the stream
    for reading. Dataset reader should be able to read the same dataset
    multiple times.
    """
    def __init__(self, reader: ArchiveReader, version: int, schema: List[Column], is_keyed: bool):
        """Initialize the archive reader.

        Parameters
        ----------
        reader: histore.archive.reader.ArchiveReader
            Reader for a dataset archive.
        version: int
            Unique version identifier for the read snapshot.
        schema: list of histore.document.schema.Column
            Schema for the read snapshot.
        is_keyed: bool
            Flag indicating whether the archive snapshots are keyed by a primary
            key or not. The type of archive key determines the value for the
            row index when reading the document.
        """
        self.reader = reader
        self.version = version
        self.colids = [c.colid for c in schema]
        self.is_keyed = is_keyed

    def close(self):
        """Close the associated archive reader and set it to None (to avoid
        repeated attempts to close it multiple times).
        """
        if self.reader is not None:
            self.reader.close()
            self.reader = None

    def next(self) -> Tuple[int, RowIndex, DataRow]:
        """Read the next row in the document.

        Returns the row position, row index and the list of cell values for each
        of the document columns. Raises a StopIteration error if an attempt is
        made to read past the end of the document.

        Returns
        -------
        tuple of int, histore.document.base.RowIndex, histore.document.base.DataRow
        """
        """Return next row from the archive reader."""
        row = self.reader.next()
        while row is not None:
            if row.timestamp.contains(self.version):
                pos, vals = row.at_version(self.version, self.colids)
                rowidx = row.rowid if self.is_keyed else row.key.value
                return pos, rowidx, vals
            row = self.reader.next()
        self.close()
        raise StopIteration()


class SnapshotReader(DefaultDocument):
    """Stream reader for rows in a dataset archive snapshot."""
    def __init__(self, reader: Callable, version: int, schema: List[Column], is_keyed: bool):
        """Initialize the function to open the archive reader and information
        about the read snapshot.

        Parameters
        ----------
        reader: callable
            Callable that returns a new archive reader.
        version: int
            Unique version identifier for the read snapshot.
        schema: list of histore.document.schema.Column
            Schema for the read snapshot.
        is_keyed: bool
            Flag indicating whether the archive snapshots are keyed by a primary
            key or not. The type of archive key determines the value for the
            row index when reading the document.
        """
        super(SnapshotReader, self).__init__(columns=schema)
        self.reader = reader
        self.schema = schema
        self.version = version
        self.is_keyed = is_keyed

    def close(self):
        """There are no resources that need to be released."""
        pass

    def open(self) -> SnapshotIterator:
        """Open the data stream to get a iterator for the rows in the dataset
        snapshot.

        Returns
        -------
        histore.archive.reader.SnapshotIterator
        """
        return SnapshotIterator(
            reader=self.reader(),
            version=self.version,
            schema=self.schema,
            is_keyed=self.is_keyed
        )


# -- Helper Methods -----------------------------------------------------------

def row_stream(reader):
    """Geterator that yields all rows in an archive.

    Parameters
    ----------
    reader: histore.arcive.reader.ArchiveReader
        Archive reader over which we are iterating.

    Returns
    -------
    histore.archive.row.ArchiveRow
    """
    while True:
        row = reader.next()
        if row is None:
            break
        yield row
